from typing import Tuple, List, Dict

from obspy.core import Stream
from obspy import UTCDateTime

from sohstationviewer.model.reftek.reftek_data import core
from sohstationviewer.model.handling_data import read_mseed_trace


def check_reftek_header(
        rt130: core.Reftek130, cur_key: Tuple[str, str],
        starttime: UTCDateTime, endtime: UTCDateTime,
        stream_header_by_key_chan: Dict[str, Dict[str, Stream]],
        cur_data_dict: Dict, cur_data_time: List[float],
        include_mp123zne: bool, include_mp456uvw: bool):
    """
    Read mseed headers of a file from the given rt130 object
        to check for time, create stream_header for retrieving gaps later.
        Requested data stream has been checked before passing to this function.

    :param rt130: RT130 object to get data stream from
    :param cur_key: Tuple of DAS serial number, experiment number of the
        current file.
    :param starttime: start of read data to skip reading actual data if not
        in range
    :param endtime: end of read data to skip reading actual data if not
        in range
    :param stream_header_by_key_chan: dict of stream header by key, chan to get
        gaps later
    :param cur_data_dict: waveform_data/mass_pos_data of the current key
    :param cur_data_time: data_time of the current key
    :param include_mp123zne: if mass position channels 1,2,3 are requested
    :param include_mp456uvw: if mass position channels 4,5,6 are requested
    """
    stream = core.Reftek130.to_stream(
        rt130,
        include_mp123=include_mp123zne,
        include_mp456=include_mp456uvw,
        headonly=True,
        verbose=False,
        sort_permuted_package_sequence=True)

    avail_trace_indexes = []
    for index, trace in enumerate(stream):
        chan_id = trace.stats['channel'].strip()
        samplerate = trace.stats['sampling_rate']
        if chan_id not in cur_data_dict:
            cur_data_dict[chan_id] = {'tracesInfo': [],
                                      'samplerate': samplerate}
        if trace.stats.npts == 0:
            #  this trace isn't available to prevent bug when creating memmap
            #  with no data
            continue
        if (starttime <= trace.stats['starttime'] <= endtime or
                starttime <= trace.stats['endtime'] <= endtime):
            avail_trace_indexes.append(index)

            if chan_id not in stream_header_by_key_chan[cur_key]:
                stream_header_by_key_chan[cur_key][chan_id] = Stream()
            stream_header_by_key_chan[cur_key][chan_id].append(trace)

            cur_data_time[0] = min(
                trace.stats['starttime'].timestamp, cur_data_time[0])
            cur_data_time[1] = max(
                trace.stats['endtime'].timestamp, cur_data_time[1])

    return avail_trace_indexes


def read_reftek_stream(
        rt130: core.Reftek130, tmp_dir: str, cur_key: Tuple[str, str],
        avail_trace_indexes: List[int], cur_data_dict: Dict,
        include_mp123zne: bool, include_mp456uvw: bool):
    """
    Read traces of a file from the given rt130 object for the index in
        avail_trace_indexes.

    :param rt130: RT130 object to get data stream from
    :param tmp_dir: dir to keep memmap files
    :param cur_key: Tuple of DAS serial number, experiment number of the
        current file.
    :param avail_trace_indexes: index of traces to get
    :param cur_data_dict: waveform_data/mass_pos_data of the current key
    :param include_mp123zne: if mass position channels 1,2,3 are requested
    :param include_mp456uvw: if mass position channels 4,5,6 are requested
    """
    # TODO: rewrite reftek to read stream with start and end time
    stream = core.Reftek130.to_stream(
        rt130,
        include_mp123=include_mp123zne,
        include_mp456=include_mp456uvw,
        headonly=False,
        verbose=False,
        sort_permuted_package_sequence=True)

    for index in avail_trace_indexes:
        trace = stream[index]
        chan_id = trace.stats['channel'].strip()
        traces_info = cur_data_dict[chan_id]['tracesInfo']
        tr = read_mseed_trace(
            trace, cur_key, chan_id, len(traces_info), tmp_dir)
        traces_info.append(tr)
