from typing import Tuple, Any

import numpy
from obspy.io.reftek.util import bcd

from sohstationviewer.model.general_data.general_record_helper import Unpacker
from sohstationviewer.model.reftek_data.reftek_reader.packet import \
    eh_et_payload_end_in_packet
from sohstationviewer.model.reftek_data.reftek_reader.packets import (
    DTExtendedHeader,
    EHETExtendedHeader, SOHExtendedHeader,
)


def decode_uncompressed(packet: bytes, data_format: str, unpacker: Unpacker
                        ) -> int:
    """
    Grab the first data point in a packet that contains uncompressed RT130 data
    (aka packets with data format 16, 32, or 33_.
    :param packet: the bytes that make up the given packet.
    :param data_format: the data format of the given packet, can be one of 16,
        32, or 33.
    :param unpacker: the unpacker to use to decode the data.
    :return: the first data point in the given packet
    """
    data = packet[24:]
    # For uncompressed RT130 data, the data format is also the size of a data
    # point in bit (aside from data format 33, which uses the same size as data
    # format 32).
    point_size = int(data_format)
    if point_size == 33:
        point_size = 32
    # Convert the size of a data point to byte because the data is stored
    # as a byte string.
    point_size = point_size // 8

    # struct.unpack uses different format characters for different point sizes.
    format_char = {2: 'h', 4: 'i'}[point_size]

    first_data_point = data[:point_size]

    return unpacker.unpack(f'{format_char}', first_data_point)[0]


def decode_compressed(packet: bytes, data_format: str, unpacker: Unpacker
                      ) -> int:
    """
    Grab the stop point in a packet that contains compressed RT130 data (aka
    packets with data format C0, C1, C2, or C3).
    We get the stop point in this case because that is what logpeek did. It
    also looks a lot better than using the start point, so that is a plus.
    :param packet: the bytes that make up the given packet.
    :param data_format: the data format of the given packet, can be one of C0,
        C1, C2, or C3. Exist only to have the same signature as
        decode_uncompressed
    :param unpacker: the unpacker to use to decode the data.
    :return: the first data point in the given packet
    """
    # The data in a compressed data packet starts at byte 64, with bytes
    # between byte 24 and 64 being fillers.
    data = packet[64:]
    first_data_point = data[8:12]
    return unpacker.unpack('i', first_data_point)[0]


def read_dt_packet(packet: bytes, unpacker: Unpacker
                   ) -> Tuple[DTExtendedHeader, Any]:
    """
    Process a DT packet and get its extended header and first data point.
    :param packet: the bytes that make up the given DT packet.
    :param unpacker: the unpacker to use to decode the data.
    :return: the extended header and first data point of the given DT packet.
    """
    decoders = {
        **dict.fromkeys(['16', '32', '33'], decode_uncompressed),
        **dict.fromkeys(['C0', 'C1', 'C2', 'C3'], decode_compressed)
    }

    event_number = int(packet[16:18].hex())
    data_stream_number = int(packet[18:19].hex())
    channel_number = int(packet[19:20].hex())
    number_of_samples = int(packet[20:22].hex())
    flags = unpacker.unpack('B', packet[22:23])[0]
    data_format = packet[23:24].hex().upper()

    extended_header = DTExtendedHeader(event_number, data_stream_number,
                                       channel_number, number_of_samples,
                                       flags, data_format)
    first_data_point = decoders[data_format](packet, data_format, unpacker)
    return extended_header, first_data_point


def read_eh_et_packet(packet: bytes, unpacker: Unpacker
                      ) -> Tuple[EHETExtendedHeader, bytes]:
    """
    Process an EH/ET packet and get its extended header and required part of
    the payload.
    :param packet: the bytes that make up the given EH/ET packet.
    :param unpacker: the unpacker to use to decode the data.
    :return: the extended header and truncated payload of the given EH/ET
        packet.
    """
    event_number = int(packet[16:18].hex())
    data_stream_number = int(packet[18:19].hex())
    flags = unpacker.unpack('B', packet[22:23])[0]
    data_format = packet[23:24].hex().upper()

    extended_header = EHETExtendedHeader(event_number, data_stream_number,
                                         flags, data_format)
    # The largest possible data point has a size of 4  bytes, so we need to
    # grab at least data.
    payload = packet[24:eh_et_payload_end_in_packet]
    return extended_header, payload


def bcd_16bit_int(_i) -> int:
    """
    Reimplement a private function of the same name in obspy. Kept here in case
    the private function is removed in a future obspy version.
    :param _i: the byte string to convert into a 16-bite integer
    :return: a 16-bit integer
    """
    _i = bcd(_i)
    return _i[0] * 100 + _i[1]


def read_soh_packet(packet: bytes, unpacker: Unpacker
                    ) -> Tuple[SOHExtendedHeader, bytes]:
    """
    Process an SOH packet and get its extended header and poyload.
    :param packet: the bytes that make up the given SOH packet.
    :param unpacker: the unpacker to use to decode the data.
    :return: the extended header and payload of the given SOH packet.
    """

    event_number = bcd_16bit_int(numpy.frombuffer(packet[16:18], numpy.uint8))
    data_stream_number = bcd(numpy.frombuffer(packet[18:19], numpy.uint8))
    channel_number = bcd(numpy.frombuffer(packet[19:20], numpy.uint8))
    number_of_samples = bcd_16bit_int(
        numpy.frombuffer(packet[20:22], numpy.uint8)
    )
    flags = unpacker.unpack('B', packet[22:23])[0]
    data_format = packet[23:24].hex().upper()

    extended_header = SOHExtendedHeader(event_number, data_stream_number,
                                        channel_number, number_of_samples,
                                        flags, data_format)
    payload = packet[24:]
    return extended_header, payload
